@@ -1,4 +1,5 @@
 import concurrent.futures
+import datetime
 import os
 import tempfile
 import typing
@@ -6,12 +7,66 @@ from abc import abstractmethod, ABC
 from typing import IO
 
 from google.protobuf.internal.encoder import _VarintBytes
+from sortedcontainers import SortedSet
 
 from bakplane.bakplane_pb2 import ErrorEntry
 from bakplane.ingestion.landing_zone import LandingZoneFactory
-from .models import Asset, Entity, Relationship, IngestionSessionContext, Error
+from .models import Asset, Entity, Relationship, IngestionSessionContext, Error, Descriptor
 
-from pybloom_live import ScalableBloomFilter
+PARTITION_NUM = os.cpu_count()
+KEY_SEPARATOR = '::'
+
+
+def reduce_pair(p1: typing.Tuple[datetime.datetime, datetime.datetime],
+                p2: typing.Tuple[datetime.datetime, datetime.datetime]):
+    a, b = p1
+    c, d = p2
+
+    if a == c:
+        if b < d:
+            return (a, d), None
+        elif b > d:
+            return (a, b), None
+        else:
+            raise RuntimeError('Not allowed')
+    elif a < c:
+        if b == d:
+            return (a, b), None
+        elif b < d:
+            if b < c:
+                return (a, b), (c, d)
+            elif b > c:
+                return (a, d), None
+            return (a, d), None
+        else:
+            return (a, b), None
+
+
+def reduce_intervals(s: SortedSet) -> SortedSet:
+    if not s:
+        raise RuntimeError('Input cannot be none.')
+
+    if len(s) == 1:
+        return s
+
+    done = SortedSet()
+
+    last = s[0]
+
+    for i in range(1, len(s)):
+        p1 = s[i]
+        i1, i2 = reduce_pair(last, p1)
+
+        if i2 is not None:
+            done.add(i1)
+            done.add(i2)
+
+            last = i2
+        else:
+            last = i1
+
+    done.add(last)
+    return done
 
 
 class Emitter(ABC):
@@ -32,6 +87,89 @@ class Emitter(ABC):
         pass
 
 
+class RelationshipReduceMap:
+    source_identity_uid: int
+    target_identity_uid: int
+    relationship_type: str
+    effective_dating: SortedSet
+
+    def __init__(self, r: Relationship):
+        self.relationship_type = r.relationship_type
+        self.target_identity_uid = r.target_identity_uid
+        self.source_identity_uid = r.source_identity_uid
+        self.effective_dating = SortedSet()
+
+        self.append(r)
+
+    def append(self, r: Relationship):
+        if r.source_identity_uid != self.source_identity_uid or r.target_identity_uid != self.target_identity_uid or r.relationship_type != self.relationship_type:
+            raise RuntimeError('Invalid relationship identity for this reducer.')
+
+        for e in r.effective_dating:
+            self.effective_dating.add((e[0], e[1]))
+
+    def reduce(self) -> Relationship:
+        reduced_intervals = []
+
+        for interval in reduce_intervals(self.effective_dating):
+            start, end = interval
+
+            reduced_intervals.append((start, end))
+
+        return Relationship(
+            relationship_type=self.relationship_type,
+            source_identity_uid=self.source_identity_uid,
+            target_identity_uid=self.target_identity_uid,
+            effective_dating=reduced_intervals
+        )
+
+
+class EntityReduceMap:
+    def __init__(self, entity: Entity):
+        self.identity = entity.identity
+        self.values = {}
+
+        self.append(entity)
+
+    def append(self, entity: Entity):
+        if entity.identity.uid != self.identity.uid:
+            raise RuntimeError(f'Entity mismatch: expected entity with identity `{entity.identity.uid}`.')
+
+        for d in entity.descriptors:
+            self.append_descriptor(d)
+
+    def append_descriptor(self, d: Descriptor):
+        key = d.attribute + KEY_SEPARATOR + d.value
+
+        if key not in self.values:
+            self.values[key] = SortedSet()
+
+        for dt in d.effective_dating:
+            self.values[key].add((dt[0], dt[1]))
+
+    def reduce(self) -> Entity:
+        descriptors: typing.List[Descriptor] = []
+
+        for key, v in self.values.items():
+            attr, val = key.split(KEY_SEPARATOR)
+
+            effective_dating = []
+
+            for interval in reduce_intervals(v):
+                start, end = interval
+
+                effective_dating.append((start, end))
+
+            descriptors.append(Descriptor(
+                attribute=attr,
+                value=val,
+                effective_dating=effective_dating
+            ))
+
+        self.values = None
+        return Entity(identity=self.identity, descriptors=descriptors)
+
+
 class DefaultEmitter(Emitter):
     def __init__(self, context: IngestionSessionContext, asset_num_sharding=8, entities_num_sharding=8,
                  rels_num_sharding=8):
@@ -58,13 +196,13 @@ class DefaultEmitter(Emitter):
             context.errors_output_location
         )
 
-        self.entities_bf = ScalableBloomFilter(mode=ScalableBloomFilter.LARGE_SET_GROWTH)
-        self.rels_bf = ScalableBloomFilter(mode=ScalableBloomFilter.LARGE_SET_GROWTH)
-
         self.asset_num_sharding = asset_num_sharding
         self.entities_num_sharding = entities_num_sharding
         self.rels_num_sharding = rels_num_sharding
 
+        self.entity_map = {}
+        self.rel_map = {}
+
     def __get_entities_fd(self, entity: Entity) -> IO:
         key = 'entities'
 
@@ -112,18 +250,10 @@ class DefaultEmitter(Emitter):
         return self.fd_map[key]
 
     def __write_entity(self, entity: Entity):
-        payload_hash = entity.payload_hash
-
-        if payload_hash in self.entities_bf:
-            return
-
-        f = self.__get_entities_fd(entity)
-        p = entity.to_proto()
-
-        f.write(_VarintBytes(p.ByteSize()))
-        f.write(p.SerializeToString())
-
-        self.entities_bf.add(payload_hash)
+        if entity.identity.uid not in self.entity_map:
+            self.entity_map[entity.identity.uid] = EntityReduceMap(entity)
+        else:
+            self.entity_map[entity.identity.uid].append(entity)
 
     def __write_error(self, error: ErrorEntry):
         f = self.__get_errors_fd(error)
@@ -133,18 +263,10 @@ class DefaultEmitter(Emitter):
         f.write(p.SerializeToString())
 
     def __write_relationship(self, relationship: Relationship):
-        payload_hash = relationship.payload_hash
-
-        if payload_hash in self.rels_bf:
-            return
-
-        f = self.__get_relationships_fd(relationship)
-        p = relationship.to_proto()
-
-        f.write(_VarintBytes(p.ByteSize()))
-        f.write(p.SerializeToString())
-
-        self.rels_bf.add(payload_hash)
+        if relationship.identity not in self.rel_map:
+            self.rel_map[relationship.identity] = RelationshipReduceMap(relationship)
+        else:
+            self.rel_map[relationship.identity].append(relationship)
 
     def __write_asset(self, asset: Asset):
         f = self.__get_assets_fd(asset)
@@ -178,8 +300,57 @@ class DefaultEmitter(Emitter):
         for a in assets:
             self.__write_asset(a)
 
+    def reduce_entities(self, identities: typing.List[int]):
+        for k in identities:
+            entity = self.entity_map[k].reduce()
+
+            f = self.__get_entities_fd(entity)
+            p = entity.to_proto()
+
+            f.write(_VarintBytes(p.ByteSize()))
+            f.write(p.SerializeToString())
+
+    def reduce_relationships(self, identities: typing.List[int]):
+        for k in identities:
+            relationship = self.rel_map[k].reduce()
+
+            f = self.__get_relationships_fd(relationship)
+            p = relationship.to_proto()
+
+            f.write(_VarintBytes(p.ByteSize()))
+            f.write(p.SerializeToString())
+
+    def __reduce_entities(self, executor: concurrent.futures.ThreadPoolExecutor):
+        pending_identities = set(self.entity_map.keys())
+        tasks = []
+
+        for i in range(PARTITION_NUM):
+            tasks.append(
+                executor.submit(self.reduce_entities, [c for c in pending_identities if c % PARTITION_NUM == i]))
+
+        for rt in tasks:
+            _ = rt.result()
+
+        self.entity_map = None
+
+    def __reduce_relationships(self, executor: concurrent.futures.ThreadPoolExecutor):
+        pending_identities = set(self.rel_map.keys())
+        tasks = []
+
+        for i in range(PARTITION_NUM):
+            tasks.append(
+                executor.submit(self.reduce_relationships, [c for c in pending_identities if c % PARTITION_NUM == i]))
+
+        for rt in tasks:
+            _ = rt.result()
+
+        self.rel_map = None
+
     def close(self):
-        with concurrent.futures.ThreadPoolExecutor(max_workers=os.cpu_count() + 1) as executor:
+        with concurrent.futures.ThreadPoolExecutor(max_workers=PARTITION_NUM) as executor:
+            self.__reduce_entities(executor)
+            self.__reduce_relationships(executor)
+
             tasks = []
 
             for k, v in self.fd_map.items():
